@@ -13,13 +13,13 @@ import operator
 import re
 import datetime
 from calendar import isleap
-from decimal import Decimal
+from decimal import Decimal, Context
 from typing import cast, Any, Callable, Dict, Optional, Tuple, Union
 
 from ..helpers import MONTH_DAYS_LEAP, MONTH_DAYS, DAYS_IN_4Y, \
     DAYS_IN_100Y, DAYS_IN_400Y, days_from_common_era, adjust_day, \
     normalized_seconds, months2days, round_number
-from .atomic_types import AtomicTypeMeta, AnyAtomicType
+from .atomic_types import AnyAtomicType
 from .untyped import UntypedAtomic
 
 
@@ -112,7 +112,7 @@ class Timezone(datetime.tzinfo):
         raise TypeError("fromutc() argument must be a datetime.datetime instance")
 
 
-class AbstractDateTime(metaclass=AtomicTypeMeta):
+class AbstractDateTime(AnyAtomicType):
     """
     A class for representing XSD date/time objects. It uses and internal datetime.datetime
     attribute and an integer attribute for processing BCE years or for years after 9999 CE.
@@ -754,7 +754,7 @@ class Duration(AnyAtomicType):
             raise OverflowError("seconds duration overflow")
 
         self.months = months
-        self.seconds = Decimal(seconds).quantize(Decimal('1.000000'))
+        self.seconds = Decimal(seconds).quantize(Decimal('1.000000', context=Context(prec=30)))
 
     def __repr__(self) -> str:
         return '{}(months={!r}, seconds={})'.format(
